'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var umiCore = require('@metaplex-foundation/umi-core');
var errors = require('./errors.cjs');
var numbers = require('./numbers.cjs');

class BeetSerializer {
  constructor(options = {}) {
    this.options = options;
  }
  tuple(items, options = {}) {
    const itemDescriptions = items.map(item => item.description).join(', ');
    return {
      description: options.description ?? `tuple(${itemDescriptions})`,
      fixedSize: sumSerializerSizes(items.map(item => item.fixedSize)),
      maxSize: sumSerializerSizes(items.map(item => item.maxSize)),
      serialize: value => {
        if (value.length !== items.length) {
          throw new errors.BeetSerializerError(`Expected tuple to have ${items.length} items but got ${value.length}.`);
        }
        return umiCore.mergeBytes(items.map((item, index) => item.serialize(value[index])));
      },
      deserialize: (bytes, offset = 0) => {
        const values = [];
        items.forEach(serializer => {
          const [newValue, newOffset] = serializer.deserialize(bytes, offset);
          values.push(newValue);
          offset = newOffset;
        });
        return [values, offset];
      }
    };
  }
  array(item, options = {}) {
    const size = options.size ?? numbers.u32();
    if (size === 'remainder' && item.fixedSize === null) {
      throw new errors.BeetSerializerError('Serializers of "remainder" size must have fixed-size items.');
    }
    return {
      description: options.description ?? `array(${item.description}; ${getSizeDescription(size)})`,
      fixedSize: getSizeFromChildren(size, [item.fixedSize]),
      maxSize: getSizeFromChildren(size, [item.maxSize]),
      serialize: value => {
        if (typeof size === 'number' && value.length !== size) {
          throw new errors.BeetSerializerError(`Expected array to have ${size} items but got ${value.length}.`);
        }
        return umiCore.mergeBytes([getSizePrefix(size, value.length), ...value.map(v => item.serialize(v))]);
      },
      deserialize: (bytes, offset = 0) => {
        if (typeof size === 'object' && bytes.slice(offset).length === 0) {
          return this.handleEmptyBuffer('array', [], offset);
        }
        const [resolvedSize, newOffset] = getResolvedSize(size, [item.fixedSize], bytes, offset);
        offset = newOffset;
        const values = [];
        for (let i = 0; i < resolvedSize; i += 1) {
          const [value, newOffset] = item.deserialize(bytes, offset);
          values.push(value);
          offset = newOffset;
        }
        return [values, offset];
      }
    };
  }
  map(key, value, options = {}) {
    const size = options.size ?? numbers.u32();
    if (size === 'remainder' && (key.fixedSize === null || value.fixedSize === null)) {
      throw new errors.BeetSerializerError('Serializers of "remainder" size must have fixed-size items.');
    }
    return {
      description: options.description ?? `map(${key.description}, ${value.description}; ${getSizeDescription(size)})`,
      fixedSize: getSizeFromChildren(size, [key.fixedSize, value.fixedSize]),
      maxSize: getSizeFromChildren(size, [key.maxSize, value.maxSize]),
      serialize: map => {
        if (typeof size === 'number' && map.size !== size) {
          throw new errors.BeetSerializerError(`Expected map to have ${size} items but got ${map.size}.`);
        }
        const itemBytes = Array.from(map, ([k, v]) => umiCore.mergeBytes([key.serialize(k), value.serialize(v)]));
        return umiCore.mergeBytes([getSizePrefix(size, map.size), ...itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const map = new Map();
        if (typeof size === 'object' && bytes.slice(offset).length === 0) {
          return this.handleEmptyBuffer('map', map, offset);
        }
        const [resolvedSize, newOffset] = getResolvedSize(size, [key.fixedSize, value.fixedSize], bytes, offset);
        offset = newOffset;
        for (let i = 0; i < resolvedSize; i += 1) {
          const [deserializedKey, kOffset] = key.deserialize(bytes, offset);
          offset = kOffset;
          const [deserializedValue, vOffset] = value.deserialize(bytes, offset);
          offset = vOffset;
          map.set(deserializedKey, deserializedValue);
        }
        return [map, offset];
      }
    };
  }
  set(item, options = {}) {
    const size = options.size ?? numbers.u32();
    if (size === 'remainder' && item.fixedSize === null) {
      throw new errors.BeetSerializerError('Serializers of "remainder" size must have fixed-size items.');
    }
    return {
      description: options.description ?? `set(${item.description}; ${getSizeDescription(size)})`,
      fixedSize: getSizeFromChildren(size, [item.fixedSize]),
      maxSize: getSizeFromChildren(size, [item.maxSize]),
      serialize: set => {
        if (typeof size === 'number' && set.size !== size) {
          throw new errors.BeetSerializerError(`Expected set to have ${size} items but got ${set.size}.`);
        }
        const itemBytes = Array.from(set, value => item.serialize(value));
        return umiCore.mergeBytes([getSizePrefix(size, set.size), ...itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        const set = new Set();
        if (typeof size === 'object' && bytes.slice(offset).length === 0) {
          return this.handleEmptyBuffer('set', set, offset);
        }
        const [resolvedSize, newOffset] = getResolvedSize(size, [item.fixedSize], bytes, offset);
        offset = newOffset;
        for (let i = 0; i < resolvedSize; i += 1) {
          const [value, newOffset] = item.deserialize(bytes, offset);
          offset = newOffset;
          set.add(value);
        }
        return [set, offset];
      }
    };
  }
  option(item, options = {}) {
    const prefix = options.prefix ?? numbers.u8();
    const fixed = options.fixed ?? false;
    let descriptionSuffix = `; ${getSizeDescription(prefix)}`;
    let fixedSize = item.fixedSize === 0 ? prefix.fixedSize : null;
    if (fixed) {
      if (item.fixedSize === null || prefix.fixedSize === null) {
        throw new errors.BeetSerializerError('Fixed options can only be used with fixed-size serializers');
      }
      descriptionSuffix += '; fixed';
      fixedSize = prefix.fixedSize + item.fixedSize;
    }
    return {
      description: options.description ?? `option(${item.description + descriptionSuffix})`,
      fixedSize,
      maxSize: sumSerializerSizes([prefix.maxSize, item.maxSize]),
      serialize: option => {
        const prefixByte = prefix.serialize(Number(umiCore.isSome(option)));
        if (fixed) {
          const itemFixedSize = item.fixedSize;
          const itemBytes = umiCore.isSome(option) ? item.serialize(option.value).slice(0, itemFixedSize) : new Uint8Array(itemFixedSize).fill(0);
          return umiCore.mergeBytes([prefixByte, itemBytes]);
        }
        const itemBytes = umiCore.isSome(option) ? item.serialize(option.value) : new Uint8Array();
        return umiCore.mergeBytes([prefixByte, itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        if (bytes.slice(offset).length === 0) {
          return this.handleEmptyBuffer('option', umiCore.none(), offset);
        }
        const fixedOffset = offset + (prefix.fixedSize ?? 0) + (item.fixedSize ?? 0);
        const [isSome, prefixOffset] = prefix.deserialize(bytes, offset);
        offset = prefixOffset;
        if (isSome === 0) {
          return [umiCore.none(), fixed ? fixedOffset : offset];
        }
        const [value, newOffset] = item.deserialize(bytes, offset);
        offset = newOffset;
        return [umiCore.some(value), fixed ? fixedOffset : offset];
      }
    };
  }
  nullable(item, options = {}) {
    const prefix = options.prefix ?? numbers.u8();
    const fixed = options.fixed ?? false;
    let descriptionSuffix = `; ${getSizeDescription(prefix)}`;
    let fixedSize = item.fixedSize === 0 ? prefix.fixedSize : null;
    if (fixed) {
      if (item.fixedSize === null || prefix.fixedSize === null) {
        throw new errors.BeetSerializerError('Fixed nullables can only be used with fixed-size serializers');
      }
      descriptionSuffix += '; fixed';
      fixedSize = prefix.fixedSize + item.fixedSize;
    }
    return {
      description: options.description ?? `nullable(${item.description + descriptionSuffix})`,
      fixedSize,
      maxSize: sumSerializerSizes([prefix.maxSize, item.maxSize]),
      serialize: option => {
        const prefixByte = prefix.serialize(Number(option !== null));
        if (fixed) {
          const itemFixedSize = item.fixedSize;
          const itemBytes = option !== null ? item.serialize(option).slice(0, itemFixedSize) : new Uint8Array(itemFixedSize).fill(0);
          return umiCore.mergeBytes([prefixByte, itemBytes]);
        }
        const itemBytes = option !== null ? item.serialize(option) : new Uint8Array();
        return umiCore.mergeBytes([prefixByte, itemBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        if (bytes.slice(offset).length === 0) {
          return this.handleEmptyBuffer('nullable', null, offset);
        }
        const fixedOffset = offset + (prefix.fixedSize ?? 0) + (item.fixedSize ?? 0);
        const [isSome, prefixOffset] = prefix.deserialize(bytes, offset);
        offset = prefixOffset;
        if (isSome === 0) {
          return [null, fixed ? fixedOffset : offset];
        }
        const [value, newOffset] = item.deserialize(bytes, offset);
        offset = newOffset;
        return [value, fixed ? fixedOffset : offset];
      }
    };
  }
  struct(fields, options = {}) {
    const fieldDescriptions = fields.map(([name, serializer]) => `${String(name)}: ${serializer.description}`).join(', ');
    return {
      description: options.description ?? `struct(${fieldDescriptions})`,
      fixedSize: sumSerializerSizes(fields.map(([, field]) => field.fixedSize)),
      maxSize: sumSerializerSizes(fields.map(([, field]) => field.maxSize)),
      serialize: struct => {
        const fieldBytes = fields.map(([key, serializer]) => serializer.serialize(struct[key]));
        return umiCore.mergeBytes(fieldBytes);
      },
      deserialize: (bytes, offset = 0) => {
        const struct = {};
        fields.forEach(([key, serializer]) => {
          const [value, newOffset] = serializer.deserialize(bytes, offset);
          offset = newOffset;
          struct[key] = value;
        });
        return [struct, offset];
      }
    };
  }
  enum(constructor, options = {}) {
    const enumValues = Object.values(constructor);
    const isNumericEnum = enumValues.some(v => typeof v === 'number');
    const valueDescriptions = enumValues.filter(v => typeof v === 'string').join(', ');
    function getVariantKeyValue(value) {
      if (typeof value === 'number') {
        return [enumValues[value], value];
      }
      const variantValue = constructor[value];
      if (typeof variantValue === 'number') {
        return [value, variantValue];
      }
      const indexOfValue = enumValues.indexOf(variantValue);
      if (indexOfValue >= 0) {
        return [variantValue, indexOfValue];
      }
      return [value, enumValues.indexOf(value)];
    }
    function checkVariantExists(variantKey) {
      if (!enumValues.includes(variantKey)) {
        throw new errors.BeetSerializerError(`Invalid enum variant. Got "${variantKey}", expected one of ` + `[${enumValues.join(', ')}]`);
      }
    }
    return {
      description: options.description ?? `enum(${valueDescriptions})`,
      fixedSize: 1,
      maxSize: 1,
      serialize: value => {
        const [variantKey, variantValue] = getVariantKeyValue(value);
        checkVariantExists(variantKey);
        return numbers.u8().serialize(variantValue);
      },
      deserialize: (bytes, offset = 0) => {
        if (bytes.slice(offset).length === 0) {
          throw new errors.DeserializingEmptyBufferError('enum');
        }
        const [value, newOffset] = numbers.u8().deserialize(bytes, offset);
        offset = newOffset;
        const [variantKey, variantValue] = getVariantKeyValue(value);
        checkVariantExists(variantKey);
        return [isNumericEnum ? variantValue : variantKey, offset];
      }
    };
  }
  dataEnum(variants, options = {}) {
    const prefix = options.prefix ?? numbers.u8();
    const fieldDescriptions = variants.map(([name, serializer]) => `${String(name)}${serializer ? `: ${serializer.description}` : ''}`).join(', ');
    const allVariantHaveTheSameFixedSize = variants.every((one, i, all) => one[1].fixedSize === all[0][1].fixedSize);
    const fixedVariantSize = allVariantHaveTheSameFixedSize ? variants[0][1].fixedSize : null;
    const maxVariantSize = maxSerializerSizes(variants.map(([, field]) => field.maxSize));
    return {
      description: options.description ?? `dataEnum(${fieldDescriptions}; ${prefix.description})`,
      fixedSize: variants.length === 0 ? prefix.fixedSize : sumSerializerSizes([prefix.fixedSize, fixedVariantSize]),
      maxSize: variants.length === 0 ? prefix.maxSize : sumSerializerSizes([prefix.maxSize, maxVariantSize]),
      serialize: variant => {
        const discriminator = variants.findIndex(([key]) => variant.__kind === key);
        if (discriminator < 0) {
          throw new errors.BeetSerializerError(`Invalid data enum variant. Got "${variant.__kind}", expected one of ` + `[${variants.map(([key]) => key).join(', ')}]`);
        }
        const variantPrefix = prefix.serialize(discriminator);
        const variantSerializer = variants[discriminator][1];
        const variantBytes = variantSerializer.serialize(variant);
        return umiCore.mergeBytes([variantPrefix, variantBytes]);
      },
      deserialize: (bytes, offset = 0) => {
        if (bytes.slice(offset).length === 0) {
          throw new errors.DeserializingEmptyBufferError('dataEnum');
        }
        const [discriminator, dOffset] = prefix.deserialize(bytes, offset);
        offset = dOffset;
        const variantField = variants[Number(discriminator)] ?? null;
        if (!variantField) {
          throw new errors.BeetSerializerError(`Data enum index "${discriminator}" is out of range. ` + `Index should be between 0 and ${variants.length - 1}.`);
        }
        const [variant, vOffset] = variantField[1].deserialize(bytes, offset);
        offset = vOffset;
        return [{
          __kind: variantField[0],
          ...(variant ?? {})
        }, offset];
      }
    };
  }
  string(options = {}) {
    const size = options.size ?? numbers.u32();
    const encoding = options.encoding ?? umiCore.utf8;
    const description = options.description ?? `string(${encoding.description}; ${getSizeDescription(size)})`;
    if (size === 'variable') {
      return {
        ...encoding,
        description
      };
    }
    if (typeof size === 'number') {
      return umiCore.fixSerializer(encoding, size, description);
    }
    return {
      description,
      fixedSize: null,
      maxSize: null,
      serialize: value => {
        const contentBytes = encoding.serialize(value);
        const lengthBytes = size.serialize(contentBytes.length);
        return umiCore.mergeBytes([lengthBytes, contentBytes]);
      },
      deserialize: (buffer, offset = 0) => {
        if (buffer.slice(offset).length === 0) {
          throw new errors.DeserializingEmptyBufferError('string');
        }
        const [length, lengthOffset] = size.deserialize(buffer, offset);
        offset = lengthOffset;
        const contentBuffer = buffer.slice(offset, offset + Number(length));
        const [value, contentOffset] = encoding.deserialize(contentBuffer);
        offset += contentOffset;
        return [value, offset];
      }
    };
  }
  bool(options = {}) {
    const size = options.size ?? numbers.u8();
    if (size.fixedSize === null) {
      throw new errors.BeetSerializerError('Serializer [bool] requires a fixed size.');
    }
    return {
      description: options.description ?? `bool(${size.description})`,
      fixedSize: size.fixedSize,
      maxSize: size.fixedSize,
      serialize: value => size.serialize(value ? 1 : 0),
      deserialize: (bytes, offset = 0) => {
        if (bytes.slice(offset).length === 0) {
          throw new errors.DeserializingEmptyBufferError('bool');
        }
        const [value, vOffset] = size.deserialize(bytes, offset);
        return [value === 1, vOffset];
      }
    };
  }
  unit(options = {}) {
    return {
      description: options.description ?? 'unit',
      fixedSize: 0,
      maxSize: 0,
      serialize: () => new Uint8Array(),
      deserialize: (_bytes, offset = 0) => [undefined, offset]
    };
  }
  u8(options = {}) {
    return numbers.u8(options);
  }
  u16(options = {}) {
    return numbers.u16(options);
  }
  u32(options = {}) {
    return numbers.u32(options);
  }
  u64(options = {}) {
    return numbers.u64(options);
  }
  u128(options = {}) {
    return numbers.u128(options);
  }
  i8(options = {}) {
    return numbers.i8(options);
  }
  i16(options = {}) {
    return numbers.i16(options);
  }
  i32(options = {}) {
    return numbers.i32(options);
  }
  i64(options = {}) {
    return numbers.i64(options);
  }
  i128(options = {}) {
    return numbers.i128(options);
  }
  f32() {
    return {
      description: 'f32 [not supported]',
      fixedSize: 4,
      maxSize: 4,
      serialize: () => {
        throw new errors.OperationNotSupportedError('f32');
      },
      deserialize: () => {
        throw new errors.OperationNotSupportedError('f32');
      }
    };
  }
  f64() {
    return {
      description: 'f64 [not supported]',
      fixedSize: 8,
      maxSize: 8,
      serialize: () => {
        throw new errors.OperationNotSupportedError('f64');
      },
      deserialize: () => {
        throw new errors.OperationNotSupportedError('f64');
      }
    };
  }
  bytes(options = {}) {
    return {
      description: options.description ?? 'bytes',
      fixedSize: null,
      maxSize: null,
      serialize: value => new Uint8Array(value),
      deserialize: (bytes, offset = 0) => [new Uint8Array(bytes), offset + bytes.length]
    };
  }
  publicKey(options = {}) {
    return {
      description: options.description ?? 'publicKey',
      fixedSize: 32,
      maxSize: 32,
      serialize: value => umiCore.publicKey(value).bytes,
      deserialize: (bytes, offset = 0) => {
        if (bytes.slice(offset).length === 0) {
          throw new errors.DeserializingEmptyBufferError('publicKey');
        }
        const pubkeyBytes = bytes.slice(offset, offset + 32);
        if (pubkeyBytes.length < 32) {
          throw new errors.BeetSerializerError(`Serializer [publicKey] expected 32 bytes, got ${pubkeyBytes.length}.`);
        }
        return [umiCore.publicKey(pubkeyBytes), offset + 32];
      }
    };
  }
  handleEmptyBuffer(serializer, defaultValue, offset) {
    if (!(this.options.tolerateEmptyBuffers ?? true)) {
      throw new errors.DeserializingEmptyBufferError(serializer);
    }
    return [defaultValue, offset];
  }
}
function sumSerializerSizes(sizes) {
  return sizes.reduce((all, size) => all === null || size === null ? null : all + size, 0);
}
function maxSerializerSizes(sizes) {
  return sizes.reduce((all, size) => all === null || size === null ? null : Math.max(all, size), 0);
}
function getSizeDescription(size) {
  return typeof size === 'object' ? size.description : `${size}`;
}
function getSizeFromChildren(size, childrenSizes) {
  if (typeof size !== 'number') return null;
  if (size === 0) return 0;
  const childrenSize = sumSerializerSizes(childrenSizes);
  return childrenSize === null ? null : childrenSize * size;
}
function getSizePrefix(size, realSize) {
  return typeof size === 'object' ? size.serialize(realSize) : new Uint8Array();
}
function getResolvedSize(size, childrenSizes, bytes, offset) {
  if (typeof size === 'number') {
    return [size, offset];
  }
  if (typeof size === 'object') {
    return size.deserialize(bytes, offset);
  }
  if (size === 'remainder') {
    const childrenSize = sumSerializerSizes(childrenSizes);
    if (childrenSize === null) {
      throw new errors.BeetSerializerError('Serializers of "remainder" size must have fixed-size items.');
    }
    const remainder = bytes.slice(offset).length;
    if (remainder % childrenSize !== 0) {
      throw new errors.BeetSerializerError(`Serializers of "remainder" size must have a remainder that is a multiple of its item size. ` + `Got ${remainder} bytes remaining and ${childrenSize} bytes per item.`);
    }
    return [remainder / childrenSize, offset];
  }
  throw new errors.BeetSerializerError(`Unknown size type: ${JSON.stringify(size)}.`);
}

exports.BeetSerializer = BeetSerializer;
//# sourceMappingURL=BeetSerializer.cjs.map
